#################################################################################
# 
#  Non-linear, robust homography estimation
#  Copyright (C) 2003-06  Manolis Lourakis (lourakis **at** .forth.gr)
#  Institute of Computer Science, Foundation for Research & Technology - Hellas
#  Heraklion, Crete, Greece.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#################################################################################

# 2D homography estimation using several cost functions; see HZ2 p. 112

with(linalg):
with(codegen):

# 9x1 vector converted to 3x3 matrix
vec2mat3x3:=proc(a::array(1..9), H::array(1..3, 1..3))
    H:=array(1..3, 1..3, [[a[1], a[2], a[3]], 
                          [a[4], a[5], a[6]],
                          [a[7], a[8], a[9]]]);
end:

# H p 
Mk2DHomogExpr:=proc(p1::array(1..3), a::array(1..9), q2::array(1..3))

local H;

    H:=array(1..3, 1..3);
    vec2mat3x3(a, H);

    q2:=array(1..3); 
    q2:=evalm(H&*p1);
end:

# Two equations per point. This assumes that the 3rd coordinate is nonzero, which is
# always true for image corners. Hence, this is suitable for estimating a homography
# from point correspondences
Mk2DHomogEqsLin2:=proc(p1::array(1..3), p2::array(1..3), a::array(1..9), eqs::array(1..2))
local q2;

    q2:=array(1..3);
    Mk2DHomogExpr(p1, a, q2);

    eqs[1]:=simplify(p2[1]*q2[3] - p2[3]*q2[1]);
    eqs[2]:=simplify(p2[2]*q2[3] - p2[3]*q2[2]);
end:

# Three equations per point. This makes no assumption about nonzero coordinates but generates
# larger linear systems. This is suitable for estimating a homography from line or
# ideal point correspondences
Mk2DHomogEqsLin3:=proc(p1::array(1..3), p2::array(1..3), a::array(1..9), eqs::array(1..3))
local q2, cp;

    q2:=array(1..3);
    Mk2DHomogExpr(p1, a, q2);

    cp:=crossprod(p2, q2); # the crossproduct must be zero
    eqs[1]:=cp[1];
    eqs[2]:=cp[2];
    eqs[3]:=cp[3];
end:

# transfer in second image: xp1=H*p1
Mk2DHomogXfer:=proc(p1::array(1..3), a::array(1..9), xp1::array(1..2))
local q2;

    q2:=array(1..3);
    Mk2DHomogExpr(p1, a, q2);
    xp1[1]:=q2[1]/q2[3];
    xp1[2]:=q2[2]/q2[3];
end:

# symmetric transfer: xp1=H*p1, xp2=H^-1*p2
Mk2DHomogSymXfer:=proc(p1::array(1..3), p2::array(1..3), a::array(1..9), xp12::array(1..4))
local q2, aH, aHinv, q1;

    # 2nd image
    q2:=array(1..3);
    Mk2DHomogExpr(p1, a, q2);
    xp12[3]:=q2[1]/q2[3];
    xp12[4]:=q2[2]/q2[3];

    # 1st image
    aH:=array(1..3, 1..3);
    vec2mat3x3(a, aH);
    aHinv:=inverse(aH);
    q1:=array(1..3);
    q1:=evalm(aHinv&*p2);
    xp12[1]:=q1[1]/q1[3];
    xp12[2]:=q1[2]/q1[3];
end:

# Sampson approximation to geometric error for a homography, see HZ2 p. 99
Mk2DHomogEqsSampson2:=proc(parms::array(1..4), e::array(1..2), err)
local tmp, i, j, Jac;

  tmp:=array(1..4);
  Jac:=array(1..2, 1..4);

  for i from 1 to 2 do
    tmp:=grad(e[i], parms);
    for j from 1 to 4 do
      Jac[i, j]:=tmp[j];
    end do:
  end do:

  err:=sqrt(evalm(transpose(e)&*inverse(Jac&*transpose(Jac))&*e));
end:

# reprojection: pp1, pp2=H*pp1
Mk2DHomogRepr:=proc(a::array(1..9), pp1::array(1..2), pp12::array(1..4))
local p1, q2, aH;

    # 1st image
    pp12[1]:=pp1[1];
    pp12[2]:=pp1[2];

    # 2nd image
    q2:=array(1..3);
    p1:=array(1..3, [pp1[1], pp1[2], 1]):
    Mk2DHomogExpr(p1, a, q2);
    pp12[3]:=q2[1]/q2[3];
    pp12[4]:=q2[2]/q2[3];
end:


# Affine homography procedures (last row = [0 0 1])

# Two equations per point for an affine H. This assumes that the 3rd coordinate is nonzero,
# which is always true for image corners. Hence, this is suitable for estimating am affine
# homography from point correspondences. Notice that in the case of an affine homography,
# the equations are no longer homogeneous!
Mk2DAffHomogEqsLin2:=proc(p1::array(1..3), p2::array(1..3), a::array(1..6), eqs::array(1..2), rhss::array(1..2))
local a9, q2, i;

    a9:=array(1..9, [a[1], a[2], a[3], a[4], a[5], a[6], 0, 0, 1]);
    q2:=array(1..3);
    Mk2DHomogExpr(p1, a9, q2);

    eqs[1]:=simplify(p2[1]*q2[3] - p2[3]*q2[1]); rhss[1]:=-eqs[1];
    eqs[2]:=simplify(p2[2]*q2[3] - p2[3]*q2[2]); rhss[2]:=-eqs[2];

    # determine the constant terms (ie RHS). Note the '-' in the initializations above
    for i from 1 to 6 do
      rhss[1]:=simplify(rhss[1]+coeff(eqs[1], a[i], 1)*a[i]);
      rhss[2]:=simplify(rhss[2]+coeff(eqs[2], a[i], 1)*a[i]);
    end:
end:




m1:=array(1..2):
m2:=array(1..2):
# non-ideal homogeneous points
p1:=array(1..3, [m1[1], m1[2], 1]):
p2:=array(1..3, [m2[1], m2[2], 1]):
meas:=array(1..4, [m1[1], m1[2], m2[1], m2[2]]):

# homogeneous points, might be ideal
n1:=array(1..3):
n2:=array(1..3):

h:=array(1..9):
haf:=array(1..6): # affine
eqs2:=array(1..2):
eqs3:=array(1..3):
eqs2aff:=array(1..2): rhsaff:=array(1..2):

Mk2DHomogEqsLin2(p1, p2, h, eqs2):
Mk2DHomogEqsLin3(n1, n2, h, eqs3):
Mk2DHomogEqsSampson2(meas, eqs2, samperr): 
Mk2DAffHomogEqsLin2(p1, p2, haf, eqs2aff, rhsaff):

xferm1:=array(1..2):
xferm12:=array(1..4):
Mk2DHomogXfer(p1, h, xferm1):
Mk2DHomogSymXfer(p1, p2, h, xferm12):

vars:=vector([h[1], h[2], h[3], 
              h[4], h[5], h[6],
              h[7], h[8], h[9]]):
varsaff:=vector([haf[1], haf[2], haf[3], 
                 haf[4], haf[5], haf[6]]):

# transfer derivatives
xferm1_jac:=array(1..2, 1..9):
xferm12_jac:=array(1..4, 1..9):
for i from 1 to 2 do
  grd:=grad(xferm1[i], vars):
  for j from 1 to 9 do
    xferm1_jac[i, j]:=grd[j]:
  end do:
end do:

for i from 1 to 4 do
  grd:=grad(xferm12[i], vars):
  for j from 1 to 9 do
    xferm12_jac[i, j]:=grd[j]:
  end do:
end do:

# Sampson derivatives
samperr_gradient:=grad(samperr, vars):

# Resprojection error derivatives
repr:=array(1..4):
pm1:=array(1..2):
Mk2DHomogRepr(h, pm1, repr):
# repr_hgrad0, repr_hgrad1 are all zero so we don't need to compute them
#repr_hgrad0:=grad(repr[1], vars):
#repr_hgrad1:=grad(repr[2], vars):
repr_hgrad2:=grad(repr[3], vars):
repr_hgrad3:=grad(repr[4], vars):
repr_pm1grad0:=grad(repr[1], [pm1[1], pm1[2]]):
repr_pm1grad1:=grad(repr[2], [pm1[1], pm1[2]]):
repr_pm1grad2:=grad(repr[3], [pm1[1], pm1[2]]):
repr_pm1grad3:=grad(repr[4], [pm1[1], pm1[2]]):

calc2DHomogLinCoeffs2:=optimize(makeproc([
        seq(eq[j]  =coeff(eqs2[1], vars[j], 1), j=1..9),\
        seq(eq[9+j]=coeff(eqs2[2], vars[j], 1), j=1..9)],\
        parameters=[m1::array(1..2), m2::array(1..2),\
        eq::array(1..2*9)])):
calc2DHomogLinCoeffs2:=makevoid(calc2DHomogLinCoeffs2):

calc2DAffHomogLinCoeffs2:=optimize(makeproc([
        seq(eq[j]  =coeff(eqs2aff[1], varsaff[j], 1), j=1..6),\
        seq(eq[6+j]=coeff(eqs2aff[2], varsaff[j], 1), j=1..6),\
        seq(rh[j]=rhsaff[j], j=1..2)],\ 
        parameters=[m1::array(1..2), m2::array(1..2),\
        eq::array(1..2*6), rh::array(1..2)])):
calc2DAffHomogLinCoeffs2:=makevoid(calc2DAffHomogLinCoeffs2):


calc2DHomogLinCoeffs3:=optimize(makeproc([
        seq(eq[j]    =coeff(eqs3[1], vars[j], 1), j=1..9),\
        seq(eq[9+j]  =coeff(eqs3[2], vars[j], 1), j=1..9),\
        seq(eq[2*9+j]=coeff(eqs3[3], vars[j], 1), j=1..9)],\
        parameters=[n1::array(1..3), n2::array(1..3),\
        eq::array(1..3*9)])):
calc2DHomogLinCoeffs3:=makevoid(calc2DHomogLinCoeffs3):


calc2DHomogTransfer:=optimize(makeproc([
                          seq(xm1[i]=xferm1[i], i=1..2)],\
                          parameters=[m1::array(1..2),\
                          h::array(1..9), xm1::array(1..2)])):
calc2DHomogTransfer:=makevoid(calc2DHomogTransfer):

calc2DHomogTransferJac:=optimize(makeproc([
                          seq(seq(xm1_jac[i, j]=xferm1_jac[i, j], j=1..9), i=1..2)],\
                          parameters=[m1::array(1..2),\
                          h::array(1..9), xm1_jac::array(1..2, 1..9)])):
calc2DHomogTransferJac:=makevoid(calc2DHomogTransferJac):


calc2DHomogSymTransfer:=optimize(makeproc([
                          seq(xm12[i]=xferm12[i], i=1..4)],\
                          parameters=[m1::array(1..2), m2::array(1..2),\
                          h::array(1..9), xm12::array(1..4)])):
calc2DHomogSymTransfer:=makevoid(calc2DHomogSymTransfer):

calc2DHomogSymTransferJac:=optimize(makeproc([
                          seq(seq(xm12_jac[i, j]=xferm12_jac[i, j], j=1..9), i=1..4)],\
                          parameters=[m1::array(1..2), m2::array(1..2),\
                          h::array(1..9), xm12_jac::array(1..4, 1..9)])):
calc2DHomogSymTransferJac:=makevoid(calc2DHomogSymTransferJac):


calc2DHomogSampsonErr:=optimize(makeproc([
                          err[1]=samperr],\
                          parameters=[m1::array(1..2), m2::array(1..2),\
                          h::array(1..9), err::array(1..1)])):
calc2DHomogSampsonErr:=makevoid(calc2DHomogSampsonErr):

calc2DHomogSampsonErrGrads:=optimize(makeproc([
                          seq(grads[i]=samperr_gradient[i], i=1..9)],\
                          parameters=[m1::array(1..2), m2::array(1..2),\
                          h::array(1..9), grads::array(1..9)])):
calc2DHomogSampsonErrGrads:=makevoid(calc2DHomogSampsonErrGrads):


calc2DHomogReprojection:=optimize(makeproc([
                          seq(pm12[i]=repr[i], i=1..4)],\
                          parameters=[h::array(1..9), pm1::array(1..2),\
                          pm12::array(1..4)])):
calc2DHomogReprojection:=makevoid(calc2DHomogReprojection):

calc2DHomogReprojectionJac:=optimize(makeproc([
                          seq(hgrad2[i]=repr_hgrad2[i], i=1..9),\
                          seq(hgrad3[i]=repr_hgrad3[i], i=1..9),\
                          seq(pm1grad0[i]=repr_pm1grad0[i], i=1..2),\
                          seq(pm1grad1[i]=repr_pm1grad1[i], i=1..2),\
                          seq(pm1grad2[i]=repr_pm1grad2[i], i=1..2),\
                          seq(pm1grad3[i]=repr_pm1grad3[i], i=1..2)],\
                          parameters=[h::array(1..9), pm1::array(1..2),\
                          hgrad2::array(1..9), hgrad3::array(1..9),\
                          pm1grad0::array(1..2), pm1grad1::array(1..2), pm1grad2::array(1..2), pm1grad3::array(1..2)])):
calc2DHomogReprojectionJac:=makevoid(calc2DHomogReprojectionJac):


# Generate C code
fname:="./calc_2Dhomog_coeffs.c";
fd:=open(fname, WRITE);
fprintf(fd, "/* This code is part of homest and was automatically generated by maple */\n\n");
fprintf(fd, "#include \"maplefuncs.h\"\n\n");
close(fd);

C(calc2DHomogLinCoeffs2, declarations='[m1::array(1..2), m2::array(1..2),\
                             eq::array(1..2*9)]', ansi, filename=fname):

C(calc2DAffHomogLinCoeffs2, declarations='[m1::array(1..2), m2::array(1..2),\
                             eq::array(1..2*6), rh::array(1..2)]', ansi, filename=fname):

C(calc2DHomogLinCoeffs3, declarations='[n1::array(1..3), n2::array(1..3),\
                             eq::array(1..3*9)]', ansi, filename=fname):

C(calc2DHomogTransfer, declarations='[m1::array(1..2), h::array(1..9),\
                                  xm1::array(1..2)]', ansi, filename=fname):

C(calc2DHomogTransferJac, declarations='[m1::array(1..2), h::array(1..9),\
                             xm1_jac::array(1..2, 1..9)]', ansi, filename=fname):

C(calc2DHomogSymTransfer, declarations='[m1::array(1..2), m2::array(1..2), h::array(1..9),\
                                  xm12::array(1..4)]', ansi, filename=fname):

C(calc2DHomogSymTransferJac, declarations='[m1::array(1..2), m2::array(1..2), h::array(1..9),\
                             xm12_jac::array(1..4, 1..9)]', ansi, filename=fname):

C(calc2DHomogSampsonErr, declarations='[m1::array(1..2), m2::array(1..2), h::array(1..9),\
                                  err::array(1..1)]', ansi, filename=fname):

C(calc2DHomogSampsonErrGrads, declarations='[m1::array(1..2), m2::array(1..2), h::array(1..9),\
                             grads::array(1..9)]', ansi, filename=fname):

C(calc2DHomogReprojection, declarations='[h::array(1..9), pm1::array(1..2),\
                                  pm12::array(1..4)]', ansi, filename=fname):

C(calc2DHomogReprojectionJac, declarations='[h::array(1..9), pm1::array(1..2),\
                             hgrad2::array(1..9), hgrad3::array(1..9),\
                             pm1grad0::array(1..2), pm1grad1::array(1..2), pm1grad2::array(1..2), pm1grad3::array(1..2)]',
                             ansi, filename=fname):
